# new version includes 4.2.0
bl_info = {
    "name": "KIT OPS Extra Utils",
    "author": "Chipp Walters, Anthony Aragues",
    "version": (1, 0, 5),
    "blender": (3, 6, 0),
    "location": "View3D > Tool Shelf > KITOPS",
    "description": "For KIT OPS: Finds duplicate images in all scene materials and remaps them to one",
    "warning": "",
    "wiki_url": "https://www.chippwalters.com",
    "category": "3D View",
}

import bpy
import os
import hashlib
import tempfile
import re # Import the regular expressions module
from mathutils import Matrix, Vector #for setting origin to bottom of object


#### REMAP MATERIAL IMAGES ####
# Ensure the temp directory exists
temp_dir = tempfile.gettempdir()
os.makedirs(temp_dir, exist_ok=True)

def save_and_hash_image(image, temp_dir, scene):
    """Save the image to a temporary file and calculate its hash."""
    temp_filename = os.path.join(temp_dir, image.name)
    temp_filepath = bpy.path.abspath(temp_filename)
    
    # Attempt to save the image to the temp directory
    try:
        # Ensure the scene parameter is correctly passed as a keyword argument
        image.save_render(filepath=temp_filepath, scene=scene)
        print(f"*** Image saved to: {temp_filepath}")
    except Exception as e:
        print(f"*** Error saving image {image.name}: {e}")
        return None

    # Calculate the hash of the saved image
    try:
        with open(temp_filepath, 'rb') as f:
            img_data = f.read()
        img_hash = hashlib.md5(img_data).hexdigest()
        print(f"*** Hash for {image.name}: {img_hash}")
    except Exception as e:
        print(f"*** Error hashing image {image.name}: {e}")
        return None

    # Optionally, clean up the temporary file
    os.remove(temp_filepath)
    print(f"*** Temporary file removed: {temp_filepath}")
    
    return img_hash


def image_hash(image, temp_dir):
    """Generate a hash for an image, handling both packed and unpacked images."""
    if image.packed_file:
        # Pass the current scene from the context to save_and_hash_image
        return save_and_hash_image(image, temp_dir, bpy.context.scene)
    else:
        # For unpacked images, directly hash the file data
        try:
            abs_path = bpy.path.abspath(image.filepath)
            with open(abs_path, 'rb') as f:
                img_data = f.read()
            img_hash = hashlib.md5(img_data).hexdigest()
            return img_hash
        except Exception as e:
            print(f"*** Error processing image {image.name}: {e}")
            return None

def check_image_nodes(node, images_info, seen_hashes, temp_dir):
    """Recursive function to check nodes for image textures, including in node groups, avoiding duplicates based on hash."""
    if node.type == 'TEX_IMAGE':
        image = node.image
        if image is not None:
            img_hash = image_hash(image, temp_dir)
            if img_hash and img_hash not in seen_hashes:
                seen_hashes.add(img_hash)  # Mark this hash as seen
                image_info = {
                    'name': image.name,
                    'filepath': image.filepath if not image.packed_file else "Packed Image",
                    'exists': os.path.exists(image.filepath) if not image.packed_file else "N/A (packed image)",
                    'packed': image.packed_file is not None,
                    'hash': img_hash
                }
                images_info.append(image_info)
    elif node.type == 'GROUP':
        # Dive into node groups
        for group_node in node.node_tree.nodes:
            check_image_nodes(group_node, images_info, seen_hashes, temp_dir)

def process_materials(temp_dir):
    """Process all materials in the scene."""
    all_images = []
    hash_to_image_info = {}  # Maps hash values to image info for detecting duplicates
    duplicates = {}  # Dictionary to hold duplicate image mappings (old image -> unique image)

    for mat in bpy.data.materials:
        if mat.use_nodes:
            images_info = []
            seen_hashes = set()  # Keep track of image hashes
            for node in mat.node_tree.nodes:
                check_image_nodes(node, images_info, seen_hashes, temp_dir)
            for image_info in images_info:
                img_hash = image_info['hash']
                if img_hash in hash_to_image_info:
                    # If the hash is already present, this is a duplicate
                    old_image_info = image_info
                    unique_image_info = hash_to_image_info[img_hash]
                    duplicates[old_image_info['name']] = unique_image_info['name']
                else:
                    # New unique image, track it
                    hash_to_image_info[img_hash] = image_info
                all_images.append(image_info)  # Track all processed images

    # Print out all unique images found in the scene 
    print("\n*** All images processed in the scene: " + str(len(duplicates)))
    for image in all_images:
        print(f"- {image['name']} (Filepath: {image['filepath']}, Packed: {image['packed']}, Hash: {image['hash']})")

    # Print duplicates
    num_replaced = 0
    if duplicates:
        
        for old_image, unique_image in duplicates.items():
            if old_image != unique_image:
                print(old_image,unique_image)
                print("\n*** Duplicate images found:")
                print(f"- {old_image} -> {unique_image}")
                # Replace all occurrences of old_image with unique_image in materials
                replace_duplicate_image(old_image, unique_image)
                num_replaced += 1
    else:
        print("*** No duplicate images found.")
    #num_replaced = 1
    
    #num_replaced -= 1
    return num_replaced

def replace_duplicate_image(old_image_name, unique_image_name):
    """Replace all occurrences of old_image_name with unique_image_name in materials."""
    print(f"*** Replacing {old_image_name} with {unique_image_name} in materials...")
    for mat in bpy.data.materials:
        if mat.use_nodes:
            for node in mat.node_tree.nodes:
                replace_image_in_node(node, old_image_name, unique_image_name)

def replace_image_in_node(node, old_image_name, unique_image_name):
    """Replace old_image_name with unique_image_name in a single node."""
    if node.type == 'TEX_IMAGE' and node.image.name == old_image_name:
        print(f"*** Found {old_image_name} in node {node.name}")
        node.image = bpy.data.images[unique_image_name]
        print(f"*** Replaced {old_image_name} with {unique_image_name} in node {node.name}")
    elif node.type == 'GROUP':
        # Recursively replace images within node groups
        for group_node in node.node_tree.nodes:
            replace_image_in_node(group_node, old_image_name, unique_image_name)

class SimpleConfirmOperator(bpy.types.Operator):
    """Simple Operator to confirm remapping."""
    bl_idname = "object.simple_confirm_operator"
    bl_label = "Remap Material Images"
    bl_description = 'Remap all binary duplicate material images to only one'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        num_replaced = process_materials(temp_dir)
        self.report({'INFO'}, f"Number of images replaced: " + str(num_replaced))
        print("Images replaced " + str(num_replaced))
        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_confirm(self, event)

#### DEL UNUSED IMAGES ####
class CleanUpUnusedMatImagesOperator(bpy.types.Operator):
    bl_idname = "wm.clean_up_unused_mat_images"
    bl_label = "Clean Up Unused Mat Images"
    bl_description = 'Delete material images with 0 users'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        # Remove unused images
        removed_images = self.remove_unused_images()

        if removed_images:
            print("Removed unused material images:")
            for image_name in removed_images:
                print(f"- {image_name}")
        else:
            print("No unused material images found.")
        
        self.report({'INFO'}, f"Number of images replaced: " + str(len(removed_images)))
        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_confirm(self, event)

    def draw(self, context):
        self.layout.label(text="Are you sure you want to clean up unused material images?")
        # Additional UI components can be added here if necessary.

    def remove_unused_images(self):
        removed_image_names = []
        images_to_remove = [image for image in bpy.data.images if not image.users]
        for image in images_to_remove:
            removed_image_names.append(image.name)
            bpy.data.images.remove(image)
        return removed_image_names
    
#### CLEAN UP DUPE MATERIALS ####
class CW_clean_duplicate_materials(bpy.types.Operator):
    bl_idname = 'wm.clean_duplicate_materials'
    bl_label = 'Clean Duplicate Materials'
    bl_description = 'Find duplicate materials by name, remap users and remove them'
    bl_options = {'REGISTER', 'UNDO'}


    def execute(self, context):
        count = len(bpy.data.materials)

        for mat in bpy.data.materials[:]:
            if re.search('[0-9][0-9][0-9]$', mat.name):
                original = mat.name[:-4]

                if original in bpy.data.materials:
                    mat.user_remap(bpy.data.materials[original])
                    bpy.data.materials.remove(mat)

        self.report({'INFO'}, F'Removed {count - len(bpy.data.materials)} materials')
        return {'FINISHED'}

#### SET ORIGIN TO BOTTOM OF ALL SELECTED OBJECTS ####
class OBJECT_OT_origin_bot(bpy.types.Operator):
    """scales OBJ origins to bot"""
    bl_idname = "object.origin_bot"
    bl_label = "Sets Obj orig to bot of obj"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'OBJECT'
    
    def execute(self, context):
        
        def origin_to_bottom(ob, matrix=Matrix()):
            me = ob.data
            mw = ob.matrix_world
            local_verts = [matrix @ Vector(v[:]) for v in ob.bound_box]
            o = sum(local_verts, Vector()) / 8
            o.z = min(v.z for v in local_verts)
            o = matrix.inverted() @ o
            me.transform(Matrix.Translation(-o))
            
            mw.translation = mw @ o
        
        for o in bpy.context.scene.objects:
            if o.type == 'MESH':
                origin_to_bottom(o)
                #origin_to_bottom(o, matrix=o.matrix_world) # global
        
        return {'FINISHED'}
 
#### TOGGLE VP DISPLAY ####
class OBJECT_OT_wire_toggle(bpy.types.Operator):
    """Toggles display mode of selected objects between Wire and Textured"""
    bl_idname = "object.wire_toggle"
    bl_label = "Display as Wire Toggle"

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'OBJECT'

    def execute(self, context):
        selected_objects = bpy.context.selected_objects
        
        if context.object.display_type == 'WIRE':
            display_type = 'BOUNDS'
            usage_type = 'EXCLUDE'
            for obj in selected_objects:
                obj.visible_camera = False
                obj.visible_diffuse = False
                obj.visible_glossy = False
                obj.visible_transmission = False
                obj.visible_volume_scatter = False
                obj.visible_shadow = False

            #bpy.context.object.ObjectLineArt.usage = 'EXCLUDE'
            
        elif context.object.display_type == 'BOUNDS':
            display_type = 'TEXTURED'
            usage_type = 'INCLUDE'
            for obj in selected_objects:
                obj.visible_camera = True
                obj.visible_diffuse = True
                obj.visible_glossy = True
                obj.visible_transmission = True
                obj.visible_volume_scatter = True
                obj.visible_shadow = True
            
#        elif context.object.display_type == 'TEXTURED':
#            display_type = 'SOLID'
        else:
            display_type = 'WIRE' 
            usage_type = 'INCLUDE'   
        
            for obj in selected_objects:
                obj.visible_camera = False
                obj.visible_diffuse = False
                obj.visible_glossy = False
                obj.visible_transmission = False
                obj.visible_volume_scatter = False
                obj.visible_shadow = False
       
        for obj in context.selected_objects: 
            obj.display_type = display_type
            try:
                obj.lineart.usage = usage_type
            except:
                print("Blender version does not support LINE ART USAGE")
                
        return {'FINISHED'}


#### ALIGN VERTS ####
## SCALE X
class OBJECT_OT_scale_x(bpy.types.Operator):
    """Ctrl+Alt+X :: scales VERTS to x=0"""
    bl_idname = "object.scale_x"
    bl_label = "Scale selected verts to X=0"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'EDIT_MESH'

    def execute(self, context):
        bpy.ops.transform.resize(value=(0, 1, 1), orient_type='GLOBAL', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), 
            orient_matrix_type='GLOBAL', constraint_axis=(True, False, False), mirror=True, use_proportional_edit=False, 
            proportional_edit_falloff='SMOOTH', proportional_size=1, use_proportional_connected=False, use_proportional_projected=False)
        
        return {'FINISHED'}

## SCALE Y
class OBJECT_OT_scale_y(bpy.types.Operator):
    """Ctrl+Alt+C :: scales VERTS to y=0"""
    bl_idname = "object.scale_y"
    bl_label = "Scale selected verts to Y=0"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'EDIT_MESH'

    def execute(self, context):
        
        bpy.ops.transform.resize(value=(1, 0, 1), orient_type='GLOBAL', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), orient_matrix_type='GLOBAL', constraint_axis=(True, False, False), mirror=True, use_proportional_edit=False, proportional_edit_falloff='SMOOTH', proportional_size=1, use_proportional_connected=False, use_proportional_projected=False)
        
        return {'FINISHED'}

## SCALE Z
class OBJECT_OT_scale_z(bpy.types.Operator):
    """Ctrl+Alt+Z :: scales VERTS to z=0"""
    bl_idname = "object.scale_z"
    bl_label = "Scale selected verts to Z=0"
    bl_options = {'INTERNAL', 'UNDO'}

    @classmethod
    def poll(cls, context):
        
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'EDIT_MESH'

    def execute(self, context):
        
        bpy.ops.transform.resize(value=(1, 1, 0), orient_type='GLOBAL', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), orient_matrix_type='GLOBAL', constraint_axis=(True, False, False), mirror=True, use_proportional_edit=False, proportional_edit_falloff='SMOOTH', proportional_size=1, use_proportional_connected=False, use_proportional_projected=False)
        
        return {'FINISHED'}

#### ALIGN VERTS IN UV EDITOR (NO BUTTON UI)
# SCALE X IN UV EDITOR
class OBJECT_OT_scale_xuv(bpy.types.Operator):
    """scales VERTS to x=0"""
    bl_idname = "object.vertscale_xuv"
    bl_label = "Scale UV selected verts to X=0"

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'EDIT_MESH'

    def execute(self, context):
        
        bpy.ops.transform.resize(value=(0, 1, 1), orient_type='VIEW', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), 
            orient_matrix_type='VIEW', constraint_axis=(True, False, False), 
            mirror=True, use_proportional_edit=False, proportional_edit_falloff='SMOOTH', 
            proportional_size=1, use_proportional_connected=False, use_proportional_projected=False)
        
        return {'FINISHED'}

# SCALE Y IN UV EDITOR
class OBJECT_OT_scale_yuv(bpy.types.Operator):
    """scales VERTS to y=0"""
    bl_idname = "object.vertscale_yuv"
    bl_label = "Scale UV selected verts to Y=0"

    @classmethod
    def poll(cls, context):
        # Checks to see if there's any active mesh object (selected or in edit mode)
        active_object = context.active_object
        return active_object is not None and active_object.type == 'MESH' and context.mode == 'EDIT_MESH'

    def execute(self, context):
        
        bpy.ops.transform.resize(value=(1, 0, 1), orient_type='VIEW', orient_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)), 
            orient_matrix_type='VIEW', constraint_axis=(True, False, False), 
            mirror=True, use_proportional_edit=False, proportional_edit_falloff='SMOOTH', 
            proportional_size=1, use_proportional_connected=False, use_proportional_projected=False)
        
        return {'FINISHED'}
    
#### KPACK CATALOG ####
class CW_store_kpacks(bpy.types.Operator):
    bl_idname = 'mesh.store_kpacks'
    bl_label = 'Get KPACKS and INSERTs'
    bl_description = 'Get KPACKs and INSERTs'

    def execute(self, context):
        bpy.ops.wm.url_open('INVOKE_DEFAULT', url='http://cw1.me/kpacks')

        return {'FINISHED'}



#### KIT OPS SUB PANEL ####
class SimplePanel(bpy.types.Panel):
    """Panel to house the Extra Utilities."""
    bl_label = "Extra Utils"
    bl_idname = "KITOPS_PT_Extra_Utils"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'KITOPS'
    bl_order = 1000
    bl_options = {'DEFAULT_CLOSED'}

    def draw(self, context):
        layout = self.layout
        
        box = layout.box()
        box.label(text="Materials Tools")
        #row = box.row()
        box.operator("wm.clean_duplicate_materials", text="Clean Up Dupe Materials")
        box.operator("object.simple_confirm_operator", text="Remap Material Images")
        box.operator("wm.clean_up_unused_mat_images", text="Delete Unused Images")
#        layout.operator("outliner.orphans_purge", text="Purge Recursive Data Blocks").do_local_ids = True, do_linked_ids = True, do_recursive = True
        op = layout.operator("outliner.orphans_purge", text="Purge Recursive Data Blocks")
        # Set properties
        op.do_local_ids = True
        op.do_linked_ids = True
        op.do_recursive = True
        
#        layout.separator()
        layout.operator("OBJECT_OT_origin_bot", text="Set Obj Origin to Bottom")
        layout.operator("OBJECT_OT_wire_toggle", text="Toggle Obj Display Type")
        render_boolean_toggle(context, layout)
        
        box = layout.box()
        box.label(text="Align Verts (edit mode only)")
        row = box.row()
        row.scale_x = 1.0
        row.operator("OBJECT_OT_scale_x", text="X")
        row.operator("OBJECT_OT_scale_y", text="Y")
        row.operator("OBJECT_OT_scale_z", text="Z")

        layout.separator()
        layout.operator("mesh.store_kpacks", text="View KPACKS Catalog")

def render_boolean_toggle(context, layout):
    row = layout.row()
    row.scale_y = 1.5
    
    obj = context.active_object
    if obj is not None:

        is_depressed = True
        my_btn = "Booleans set to EXACT"
        for modifier in obj.modifiers:
            if modifier.type == "BOOLEAN":
                if modifier.solver == "FAST":
                    my_btn = "Booleans set to FAST"
                    is_depressed = False
                    pass
        
        
        row.operator("mesh.ko_bool_solver_toggle", text=my_btn, depress=is_depressed) 


#### SET KEYMAPS ####

addon_keymaps = []
def registerKeymaps():
    wm = bpy.context.window_manager
    if wm.keyconfigs.addon:
        # KEY SHORTCUTS FOR SPACE TYPE VIEW_3D
        km = wm.keyconfigs.addon.keymaps.new(name='3D View Generic', space_type='VIEW_3D')
        
        ## SET KEY BINDING BELOW
        ## OBJECT WIRE TOGGLE
        kmi = km.keymap_items.new('object.wire_toggle', 'Z', 'PRESS', shift=True, alt=True,ctrl=True)
        addon_keymaps.append((km, kmi))
                
        ## SCALE VERTS X
        kmi = km.keymap_items.new('object.scale_x', 'X', 'PRESS', alt=True,ctrl=True)
        addon_keymaps.append((km, kmi))

        ## SCALE VERTS Y
        kmi = km.keymap_items.new('object.scale_y', 'C', 'PRESS', alt=True,ctrl=True)
        addon_keymaps.append((km, kmi))

        ## SCALE VERTS Z
        kmi = km.keymap_items.new('object.scale_z', 'Z', 'PRESS', alt=True,ctrl=True)
        addon_keymaps.append((km, kmi))

        ## SET BOOL CUTTERS TO TARGET SLOT 0 MATERIAL
#        kmi = km.keymap_items.new('object.fix_bool_mats', 'B', 'PRESS', shift=True, alt=True,ctrl=True)
#        addon_keymaps.append((km, kmi))
        
        ## SET ORIG TO BOT
        kmi = km.keymap_items.new('object.origin_bot', 'B', 'PRESS', shift=True, alt=True,ctrl=True)
        addon_keymaps.append((km, kmi))        
        
        # KEY SHORTCUTS FOR SPACE TYPE UV
        km_uv = wm.keyconfigs.addon.keymaps.new(name='Image', space_type='IMAGE_EDITOR')
        
        ## SCALE UV VERTS X
        km_uv_i = km_uv.keymap_items.new('object.vertscale_xuv', 'X', 'PRESS', alt=True,ctrl=True)
        addon_keymaps.append((km_uv, km_uv_i))

        ## SCALE UV VERTS Y
        km_uv_i = km_uv.keymap_items.new('object.vertscale_yuv', 'C', 'PRESS', alt=True,ctrl=True)
        addon_keymaps.append((km_uv, km_uv_i))


def unregisterKeymaps():
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()

       
classes = (
    CW_clean_duplicate_materials,
    SimpleConfirmOperator,
    SimplePanel,
    CleanUpUnusedMatImagesOperator,
    OBJECT_OT_origin_bot,
    OBJECT_OT_wire_toggle,
    OBJECT_OT_scale_x,
    OBJECT_OT_scale_y,
    OBJECT_OT_scale_z,
    OBJECT_OT_scale_xuv,
    OBJECT_OT_scale_yuv,
    CW_store_kpacks,
)

def register():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)
    registerKeymaps()

def unregister():
    unregisterKeymaps()
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)
            
# def register():
#    bpy.utils.register_class(SimpleConfirmOperator)
#    bpy.utils.register_class(SimplePanel)
#    bpy.utils.register_class(CleanUpUnusedMatImagesOperator)

# def unregister():
#    bpy.utils.unregister_class(SimpleConfirmOperator)
#    bpy.utils.unregister_class(SimplePanel)
#    bpy.utils.unregister_class(CleanUpUnusedMatImagesOperator)

if __name__ == "__main__":
    register()
